import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.LinkedList;

class Test {
    static Class<?> addrCls;
    static Constructor<?> addrCtr;
    static Method addrFromString;
    static Method addrSame;
    static Method insideRng;
    static Method addrToString;

    static Class<?> pktCls;
    static Constructor<?> pktCtr;
    static Method pktGetDest;
    static Method recordPath;
    static Method pktToStr;
    static Method isLive;

    static Class<?> nodeCls;
    static Constructor<?> rootCtr;
    static Constructor<?> nodeCtr;
    static Method tracePkt;

    static Object ip1 = null;
    static Object ip2 = null;

    public static void main(String[] args) throws Exception {
        try {
            System.out.println("Az IPAddress osztaly ellenorzese:");
            runIPAddressTests();
            System.out.println("...sikeres!");
            System.out.println("A Packet osztaly ellenorzese:");
            runPacketTests();
            System.out.println("...sikeres!");
            System.out.println("A Node osztaly ellenorzese:");
            runNodeTests();
            System.out.println("...sikeres!");
        } catch(InvocationTargetException e) {
            System.err.println("Varatlan kivetel lepett fel az ellenorzes soran: " + e.getCause() + "!");
            StackTraceElement[] elements = e.getCause().getStackTrace();
            for (int i=0; i<elements.length && userSTE(elements[i]); i++) {
                System.err.println(elements[i].toString());
            }
            System.exit(1);
        }
        System.out.println("Az interfesz ellenorzese:");
        checkInterface(addrCls, Arrays.asList( (Object) addrCtr, addrFromString, addrSame, insideRng, addrToString ));
        checkInterface(pktCls, Arrays.asList( (Object) pktCtr, pktGetDest, recordPath, pktToStr, isLive ));
        checkInterface(nodeCls, Arrays.asList( (Object) rootCtr, nodeCtr, tracePkt ));
        System.out.println("...sikeres!");
    }

    public static boolean userSTE(StackTraceElement ste) {
        return ste.getClassName().startsWith("router");
    }

    public static void runIPAddressTests() throws Exception {
        addrCls = loadClass("router.IPAddress");
        addrCtr = loadCtor(addrCls, int[].class);
        addrFromString = loadStaticMethod(addrCls, "fromString", String.class);
        addrSame = loadInstanceMethod(addrCls, "isTheSame", addrCls);
        insideRng = loadInstanceMethod(addrCls, "insideRng", addrCls, addrCls);
        addrToString = loadInstanceMethod(addrCls, "toString");

        ip1 = addrCtr.newInstance(new int[] {36,43,76,100});
        ip2 = addrCtr.newInstance(new int[] {36,43,76,99});
        Object ip3 = addrFromString.invoke(null, "36.43.76.100");
        Object ip4 = addrFromString.invoke(null, "36.43.76.80");
        Object ip5 = addrFromString.invoke(null, "36.43.76.100");
        Object ip6 = addrFromString.invoke(null, "37.43.76.80");

        assertFalse( "Ket kulonbozo IP-cim eseten az isTheSame() metodus true ertekkel tert vissza!"
                   , addrSame.invoke(ip1, ip2) );
        assertFalse( "Ket kulonbozo IP-cim eseten az isTheSame() metodus true ertekkel tert vissza!"
                   , addrSame.invoke(ip2, ip1) );
        assertTrue( "Azonos objektumra hivva az isTheSame() metodus nem true ertekkel tert vissza!"
                  , addrSame.invoke(ip1,ip1) );
        assertTrue( "Megegyezo objektumra hivva az isTheSame() metodus nem true ertekkel tert vissza!"
                  , addrSame.invoke(ip1,ip3) );

        String insideRngTrueForOut = "Tartomanyon kivuli cim eseten az insideRng() metodus true ertekkel tert vissza!";
        String insideRngFalseForIn = "Tartomanyon beluli cim eseten az insideRng() metodus false ertekkel tert vissza!";

        assertFalse( insideRngTrueForOut, insideRng.invoke(ip4, ip1, ip2) );
        assertFalse( insideRngTrueForOut, insideRng.invoke(ip1, ip2, ip2) );
        assertFalse( insideRngTrueForOut, insideRng.invoke(ip4, ip5, ip6) );
        
	assertTrue( insideRngFalseForIn, insideRng.invoke(ip1, ip1, ip1) );
        assertTrue( insideRngFalseForIn, insideRng.invoke(ip1, ip2, ip1) );
        assertTrue( insideRngFalseForIn, insideRng.invoke(ip3, ip2, ip1) );
        assertTrue( insideRngFalseForIn, insideRng.invoke(ip5, ip4, ip6) );

        String toStrFormat = "Az IP-cim toString() metodusa nem megfelelo formatumú szoveget adott vissza!";
        assertEquals( toStrFormat, "36.43.76.100", addrToString.invoke(ip1) );
        assertEquals( toStrFormat, "36.43.76.80", addrToString.invoke(ip4) );

    }

    public static void runPacketTests() throws Exception {
        pktCls = loadClass("router.Packet"); 
        pktCtr = loadCtor(pktCls, String.class, addrCls);
        pktGetDest = loadInstanceMethod(pktCls, "getDestination");
        recordPath = loadInstanceMethod(pktCls, "recordPath", addrCls);
        pktToStr = loadInstanceMethod(pktCls, "toString");
        isLive = loadInstanceMethod(pktCls, "isLive");

        Object pkt = pktCtr.newInstance("X", ip1);
        recordPath.invoke(pkt, ip2);
        assertTrue( "A csomag isLive() metodusa nem adott igazat 1 ugras utan!", isLive.invoke(pkt));
        assertTrue( "A csomag lekerdezett celja nem egyezik a megadott cellal!", addrSame.invoke(pktGetDest.invoke(pkt), ip1));

        Object res = pktToStr.invoke(pkt);
        assertEquals( "A Packet toString() metodusa nem a megfelelo formaban adta vissza a csomag útjat es tartalmat!"
                    , res, "36.43.76.99 , X" );
    }

    public static void runNodeTests() throws Exception {
        nodeCls = loadClass("router.Node");
        rootCtr = loadCtor(nodeCls, addrCls, addrCls, addrCls);
        nodeCtr = loadCtor(nodeCls, nodeCls, addrCls, addrCls, addrCls);
        tracePkt = loadInstanceMethod(nodeCls, "tracePacket", pktCls);

        Object ipa1 = addrFromString.invoke(null, "0.0.0.0");
        Object ipa2 = addrFromString.invoke(null, "0.0.0.1");
        Object ipa3 = addrFromString.invoke(null, "0.0.0.2");
        Object ipa4 = addrFromString.invoke(null, "22.127.127.127");
        Object ipa5 = addrFromString.invoke(null, "23.0.0.0");
        Object ipa6 = addrFromString.invoke(null, "63.127.127.127");
        Object ipa7 = addrFromString.invoke(null, "64.0.0.0");
        Object ipa7a = addrFromString.invoke(null, "64.0.0.1");
        Object ipa8 = addrFromString.invoke(null, "126.127.127.127");
        Object ipa9 = addrFromString.invoke(null, "127.0.0.0");


        Object root     = rootCtr.newInstance(ipa1, ipa1, ipa8);
        Object node1    = nodeCtr.newInstance(root, ipa2, ipa2, ipa6);
        Object node1_1  = nodeCtr.newInstance(node1, ipa3, ipa3, ipa4);
        Object node1_2  = nodeCtr.newInstance(node1, ipa5, ipa5, ipa6);
        Object node2    = nodeCtr.newInstance(root, ipa7, ipa7, ipa8);

        Object pkt1 = pktCtr.newInstance("A", ipa5);
        Object pkt2 = pktCtr.newInstance("B", ipa3);
        Object pkt3 = pktCtr.newInstance("C", ipa2);
        Object pkt4 = pktCtr.newInstance("D", ipa1);
        Object pkt5 = pktCtr.newInstance("E", ipa9);
        Object pkt6 = pktCtr.newInstance("F", ipa7a);

        String notArrived = "Egy elkuldott csomag nem a megfelelo allomasokon jart vegig!";

        assertEquals( notArrived, tracePkt.invoke(node2, pkt1) , "64.0.0.0 , 0.0.0.0 , 0.0.0.1 , 23.0.0.0 , A");
        assertEquals( notArrived, tracePkt.invoke(node2, pkt2) , "64.0.0.0 , 0.0.0.0 , 0.0.0.1 , 0.0.0.2 , B");
        assertEquals( notArrived, tracePkt.invoke(node2, pkt3) , "64.0.0.0 , 0.0.0.0 , 0.0.0.1 , C");
        assertEquals( notArrived, tracePkt.invoke(node1_1, pkt4) , "0.0.0.2 , 0.0.0.1 , 0.0.0.0 , D");
        assertEquals( notArrived, tracePkt.invoke(node1_2, pkt5) , "OUT");
        assertEquals( notArrived, tracePkt.invoke(node1, pkt6) , "LOOP");
    }

    public static void assertTrue(String msg, Object b) {
        if (!(Boolean) b) {
            System.err.println("HIBA: " + msg);
            System.exit(1);
        }
    }

    public static void assertFalse(String msg, Object b) {
        if ((Boolean) b) {
            System.err.println("HIBA: " + msg);
            System.exit(1);
        }
    }

    public static void assertEquals(String msg, Object a, Object b) {
        if (!a.equals(b)) {
            System.err.println("HIBA: " + msg + ": '" + a.toString() + "' /= '" + b.toString() + "'");
            System.exit(1);
        }
    }

    public static void checkInterface(Class<?> tested, List<Object> methods) {
        List<Object> remMethods = new LinkedList<Object>();
        remMethods.addAll(Arrays.asList(tested.getMethods()));
        remMethods.addAll(Arrays.asList(tested.getConstructors()));
        remMethods.removeAll(methods);
        remMethods.removeAll(Arrays.asList(Object.class.getMethods()));
        if(!remMethods.isEmpty()) {
            System.err.println(String.format("HIBA: A(z) %s osztaly olyan publikus tagot tartalmaz, amelyet nem kert a feladat: '%s'!", tested.getName(), remMethods.toString()));
            System.exit(1);
        }
    }

    public static Class<?> loadClass(String className) throws Exception {
        try {
            return Class.forName(className);
        } catch(ClassNotFoundException e) {
            System.err.println("Nem talalhato a " + className + " osztaly");
            System.exit(1);
            return null;
        }
    }

    public static Constructor<?> loadCtor(Class<?> classToCtr, Class<?>... arguments) throws Exception {
        try {
            return classToCtr.getDeclaredConstructor(arguments);
        } catch(NoSuchMethodException e) {
            System.err.println("HIBA: Nem talalhato a(z) " + classToCtr + " konstruktora, amely a " + Arrays.asList(arguments).toString() + " parametereket fogadja el!" );
            System.exit(1);
            return null;
        }
    }

    public static Method loadMethod(Class<?> methodClass, String methodName, Class<?>... arguments) throws Exception {
        try {
            return methodClass.getDeclaredMethod(methodName,arguments);
        } catch(NoSuchMethodException e) {
            System.err.println("HIBA: Nem talalhato a(z) " + methodClass + " osztalyon beluli " + methodName + " metodus, amely a " + Arrays.asList(arguments).toString() + " parametereket fogadja el." );
            System.exit(1);
            return null;
        }
    }
	
	public static Method loadStaticMethod(Class<?> methodClass, String methodName, Class<?>... arguments) throws Exception {
        Method ret = loadMethod(methodClass, methodName, arguments);
		if (Modifier.isStatic(ret.getModifiers())) {
			return ret;
		} else {
			System.err.println("HIBA: A(z) " + methodClass + " osztalyon beluli " + methodName + " metodus nem statikus." );
            System.exit(1);
            return null;
		}
    }
	
	public static Method loadInstanceMethod(Class<?> methodClass, String methodName, Class<?>... arguments) throws Exception {
        Method ret = loadMethod(methodClass, methodName, arguments);
		if (!Modifier.isStatic(ret.getModifiers())) {
			return ret;
		} else {
			System.err.println("HIBA: A(z) " + methodClass + " osztalyon beluli " + methodName + " metodus statikus." );
            System.exit(1);
            return null;
		}
    }
}
