package router;

import java.util.*;

public class Node {

    private Node parent;
    private List<Node> children = new LinkedList<Node>();
    private IPAddress selfIP;
    private IPAddress lowerBound;
    private IPAddress upperBound;
    
    public Node(IPAddress ip, IPAddress lowB, IPAddress upB) {
        selfIP = ip;
        lowerBound = lowB;
        upperBound = upB;
    }
    
    public Node(Node parent, IPAddress ip, IPAddress lowB, IPAddress upB) {
        this(ip, lowB, upB);
	this.parent = parent;
	parent.children.add(this);
    }
    
    private boolean insideRng(IPAddress ip) {
        return ip.insideRng(lowerBound, upperBound);
    }
    
    public String tracePacket(Packet pkt) {
        pkt.recordPath(selfIP);
	if (selfIP.isTheSame(pkt.getDestination())) {
	    return pkt.toString();
	}
	if (! pkt.isLive()) {
	    return "LOOP";
	}
	for (Node n : children) {
	    if (n.insideRng(pkt.getDestination())) {
		return n.tracePacket(pkt);
	    }
	}
	if (parent == null) {
	    return "OUT";
	} else {
	    return parent.tracePacket(pkt);
	}
    }
}
